// Copyright (c) 1999-2014  Pavel Rychly, Milos Jakubicek

#include <finlib/config.hh>
#include <finlib/consumer.hh>
#include <finlib/srtruns.hh>
#include <finlib/bitio.hh>
#include <finlib/bititer.hh>
#include <finlib/revidx.hh>
#include <finlib/fromtof.hh>
#include <vector>
#include <limits>


using namespace std;

static void write_signature (OutFileBits_tell &ofb)
{
    static const char drevsignature[7] = "\243finDR";
    for (const char *c=drevsignature; *c; c++, ++ofb)
        *ofb = *c;
    *ofb=0;
}


typedef pair<int,Position> idpos;

static string numbered_name (const string &base, int num)
{
    char s[8];
    sprintf (s, "#%i", num);
    return base + s;
}

static void rename_rev_files (string frombase, string tobase)
{
    rename((frombase + ".rev.cnt").c_str(), (tobase + ".rev.cnt").c_str()); 
    rename((frombase + ".rev.cnt64").c_str(), (tobase + ".rev.cnt64").c_str());
    rename((frombase + ".rev.idx").c_str(), (tobase + ".rev.idx").c_str()); 
    rename((frombase + ".rev").c_str(),     (tobase + ".rev").c_str()); 
}

typedef write_bits<OutFileBits_tell&,unsigned char, OutFileBits_tell&, 
                   Position> PosWriter;

class tempdeltarev {
    struct uniq_data {
        int filenum;
        idpos prev;
        ToFile<uint32_t> *cntf, *idxf;
        ToFile<int64_t> *cntf64;
        FILE *revf;
        OutFileBits_tell *fiter;
        PosWriter *crevs;
        int nextid;
        NumOfPos id_count;
        uniq_data () :filenum (0), prev (-1,0), cntf (NULL), idxf(NULL),
            cntf64(NULL), nextid (0), id_count (0) {}
    };
    uniq_data *data;
    const string filename;
    const int alignmult;
    bool origin;

    void open_next ();
    void write_id (int newid);

public:
    int num_of_files() const {return data->filenum;}
    tempdeltarev (tempdeltarev &c) 
        : data (c.data), filename (c.filename), alignmult (c.alignmult), 
          origin (false) {}
    tempdeltarev (const char *filebase, int align, bool append=false)
        : data (new uniq_data), filename (filebase), alignmult (align), 
          origin (true)
    { 
        if (append) {
            rename_rev_files (filename, numbered_name (filename, 0));
            data->filenum++;
        }
        open_next(); 
    }
    ~tempdeltarev () {
        if (origin) {
            write_id (data->prev.first);
            delete data->crevs; delete data->fiter; delete data->cntf64;
            delete data->cntf; fclose (data->revf); delete data->idxf;
        }
    }
    void operator() (idpos x);    
    void nextRun() {}
    off_t get_curr_seek();
};

void tempdeltarev::open_next () 
{
    if (data->cntf) {
        delete data->crevs; delete data->fiter; delete data->cntf64;
        delete data->cntf; fclose (data->revf); delete data->idxf;
    }
    string s = numbered_name (filename, data->filenum);
    data->cntf = new ToFile<uint32_t>(s + ".rev.cnt");
    data->cntf64 = new ToFile<int64_t>(s + ".rev.cnt64");
    data->idxf = new ToFile<uint32_t>(s + ".rev.idx");
    data->revf = fopen ((s + ".rev").c_str(), "wb");
    data->filenum ++;
    data->fiter = new OutFileBits_tell (data->revf);
    data->crevs = new PosWriter (*data->fiter);
    write_signature (*data->fiter);
    data->crevs->delta (alignmult +1);
}

void tempdeltarev::operator() (idpos x)
{
    if (data->prev.first != x.first) {
        write_id (x.first);
        data->id_count = 1;
        data->crevs->delta (x.second +1);
    } else if (x.second != data->prev.second) {
        data->id_count++;
        data->crevs->delta (x.second - data->prev.second);
    }
    data->prev = x;
}

void tempdeltarev::write_id (int newid)
{
    if (data->prev.first != -1) {
        if (data->id_count > numeric_limits<uint32_t>::max()) {
            data->cntf64->put (data->prev.first);
            data->cntf64->put (data->id_count);
            data->cntf->put (0);
        } else {
            data->cntf->put (data->id_count);
        }
    }

    data->crevs->new_block();
    off_t seek = get_curr_seek();

    data->idxf->put (seek);
    if (data->prev.first > newid || seek > numeric_limits<uint32_t>::max()) {
        open_next();
        data->nextid = 0;
        data->crevs->new_block();
        seek = get_curr_seek();
        data->idxf->put (seek);
    }
    while (data->nextid++ < newid) {
        data->cntf->put (0);
        data->idxf->put (seek);
    }
}

off_t tempdeltarev::get_curr_seek()
{
    off_t seeko = data->fiter->tell();
    while (seeko % alignmult) {
        *++(*data->fiter) = 0;
        ++seeko;
    }
    return seeko / alignmult;
}

//message msg;

typedef delta_revidx<BinCachedFile<uint8_t,64>,
                     BinCachedFile<uint32_t,32> > cfile_cfile_delta_revidx;
typedef vector<cfile_cfile_delta_revidx*> Revs;

void revs_reservation (const string &outfilename, bool open) {
    static FILE *revf = NULL, *cntf = NULL, *cntf64 = NULL, *idxf = NULL;
    if (open) {
        revf = fopen ((outfilename + ".rev").c_str(), "wb");
        cntf = fopen ((outfilename + ".rev.cnt").c_str(), "wb");
        cntf64 = fopen ((outfilename + ".rev.cnt64").c_str(), "wb");
        idxf = fopen ((outfilename + ".rev.idx").c_str(), "wb");
    } else {
        fclose (revf);
        fclose (cntf);
        fclose (cntf64);
        fclose (idxf);
    }
}
#define RESERVE_REVS(x) revs_reservation(x, true)
#define UNRESERVE_REVS(x) revs_reservation(x, false)

/*
 * Assumes that positions are sorted only within a single run,
 * not across all runs
 */
static void join_id_part_sort (int id, Revs &inrevs, PosWriter &crevs,
                               ToFile<uint32_t> &cntf, ToFile<int64_t> &cntf64)
{
    vector<FastStream*> *fsv = new vector<FastStream*>(inrevs.size());
    NumOfPos total = 0;
    Position lastpos = -1;
    Revs::iterator f = inrevs.begin();
    int i = 0;
    while (f != inrevs.end()) {
        if ((*f)->maxid() <= id) {
            delete *f;
            f = inrevs.erase (f);
            fsv->pop_back();
            continue;
        }
        NumOfPos count = (*f)->count (id);
        total += count;
        (*fsv)[i++] = (*f)->id2poss (id);
        f++;
    }
    if (fsv->size()) {
        QOrVNode *fss = new QOrVNode (fsv);
        while (fss->peek() < fss->final()) {
            Position d = fss->next();
            if (d > lastpos)
                crevs.delta (d - lastpos);
            else
                total--;
            lastpos = d;
        }
        delete fss;
        if (total > numeric_limits<uint32_t>::max()) {
            cntf64 (id);
            cntf64 (total);
            cntf (0);
        } else {
            cntf (total);
        }
    } else
        delete fsv;
}

/*
 * Assumes that positions are sorted within each run and across all runs, i.e.
 * all positions in run #i are smaller than all positions in runs #i+1,...,#n
 */
static void join_id_total_sort (int id, Revs &inrevs, PosWriter &crevs,
                               ToFile<uint32_t> &cntf, ToFile<int64_t> &cntf64)
{
    NumOfPos total = 0;
    Position lastpos = -1;
    Revs::iterator f=inrevs.begin();
    while (f != inrevs.end()) {
        if ((*f)->maxid() <= id) {
            delete *f;
            f = inrevs.erase (f);
            continue;
        }
        NumOfPos count = (*f)->count (id);
        total += count;

        FastStream *s = (*f)->id2poss (id);
        while (count--) {
            Position d = s->next();
            if (d > lastpos)
                crevs.delta (d - lastpos);
            else
                total--;
            lastpos = d;
        }
        delete s;
        f++;
    }
    if (inrevs.size()) {
        if (total > numeric_limits<uint32_t>::max()) {
            cntf64 (id);
            cntf64 (total);
            cntf (0);
        } else {
            cntf (total);
        }
    }
}

static void joinrevfiles (Revs &inrevs, const string &outfilename, 
                          int alignmult, bool partially_sorted)
{
    {
        // test file size of the result 
        long long totaldatasize = 0;
        for (Revs::iterator f=inrevs.begin(); f != inrevs.end(); f++)
            totaldatasize += (*f)->data_size();
        if (totaldatasize > ((long long) alignmult) * 2147483647) {
            alignmult = totaldatasize / 2147483647 +1;
        }
    }
    FILE *revf = fopen ((outfilename + ".rev").c_str(), "wb");
    ToFile<uint32_t> cntf (outfilename + ".rev.cnt");
    ToFile<int64_t> cntf64 (outfilename + ".rev.cnt64");
    ToFile<uint32_t> idxf (outfilename + ".rev.idx");
    {
        //XXX musi tady byt vlozeny blok, aby se volal destruktor pred fclose
        OutFileBits_tell _xxx (revf);
        write_signature (_xxx);
        PosWriter crevs (_xxx);
        crevs.delta (alignmult +1);
        //msg ("join init OK");
        int id = 0;
        while (inrevs.size()) {
            crevs.new_block();
            off_t seeko = _xxx.tell();
            while (seeko % alignmult) {
                *++_xxx = 0;
                ++seeko;
            }
            idxf (seeko / alignmult);
            if (partially_sorted)
                join_id_part_sort (id, inrevs, crevs, cntf, cntf64);
            else
                join_id_total_sort (id, inrevs, crevs, cntf, cntf64);
            id++;
        }
        //XXX tady se vola destruktor _xxx
    }
    fclose (revf);
}



void finish_rev_files (const string &outname, int num_of_files, int align,
                       bool part_sorted)
{
    if (num_of_files > 1) {
        // join parts
        Revs r;
        int from_num = 0;
        while (from_num < num_of_files) {
            RESERVE_REVS (outname);
            r.clear();
            if (from_num > 0) {
                r.push_back (new cfile_cfile_delta_revidx
                             (numbered_name (outname, 0)));
            }
            bool finished = true;
            for (int i = from_num; i < num_of_files; i++) {
                try {
                    r.push_back (new cfile_cfile_delta_revidx
                                 (numbered_name (outname, i))); 
                } catch (FileAccessError &e) {
                    if ((e.err == EMFILE || e.err == ENFILE)
                        && i > from_num + 1) {
                        delete r.back();
                        r.erase (r.end() -1);
                        delete r.back();
                        r.erase (r.end() -1);
                        //cerr << "partial join " << from_num << '-' << i-2 << endl;
                        UNRESERVE_REVS (outname);
                        joinrevfiles (r, outname, align, part_sorted);
                        rename_rev_files (outname, numbered_name (outname, 0));
                        from_num = i -2;
                        finished = false;
                        break;
                    } else {
                        throw e;
                    }
                }
            }
            if (finished) break;
        }
        UNRESERVE_REVS (outname);
        joinrevfiles (r, outname, align, part_sorted);
        for (int i = 0; i < num_of_files; i++) {
            string s = numbered_name (outname, i);
            unlink ((s + ".rev").c_str());
            unlink ((s + ".rev.cnt").c_str());
            unlink ((s + ".rev.cnt64").c_str());
            unlink ((s + ".rev.idx").c_str());
        }
    } else {
        // rename the only part
        rename_rev_files (numbered_name (outname, 0), outname);
    }
}

class DeltaRevFileConsumer : public RevFileConsumer
{
    tempdeltarev *oi;
    SortedRuns<tempdeltarev, idpos> *drevcons;
    const string outname;
    const int alignmult;
    bool part_sorted;
public:
    DeltaRevFileConsumer (const char *filename, int buff_size, int align,
                          bool append=false, bool part_sort=false)
        : oi (new tempdeltarev (filename, align, append)), 
          drevcons (new SortedRuns<tempdeltarev, idpos> (oi, buff_size)),
          outname (filename), alignmult (align), part_sorted (part_sort)
    {}
    virtual ~DeltaRevFileConsumer() {
        drevcons->flush();
        int num_of_files = oi->num_of_files();
        delete drevcons;
        finish_rev_files (outname, num_of_files, alignmult, part_sorted);
    }
    virtual void put (int id, Position pos) {
        (*drevcons) (idpos(id, pos));
    }
};

RevFileConsumer* 
RevFileConsumer::create (const string &path, int buff_size, int align,
                         bool append, bool part_sort)
{
    return new DeltaRevFileConsumer
           (path.c_str(), buff_size, align, append, part_sort);
}

// vim: ts=4 sw=4 sta et sts=4 si cindent tw=80:
